"""Tests for `asynccpu` package."""
import _thread
import asyncio
import multiprocessing
import os
import signal
import sys
import threading
from asyncio.exceptions import CancelledError
from asyncio.futures import Future
from concurrent.futures.process import ProcessPoolExecutor
from logging import getLogger
from multiprocessing.context import Process
from subprocess import Popen
from typing import Any, Callable, List, cast

import psutil
import pytest

from asynccpu import ProcessTaskPoolExecutor
from asynccpu.process_task_pool_executor import run, terminate_processes
from tests.testlibraries import (
    SECOND_SLEEP_FOR_TEST_KEYBOARD_INTERRUPT_CTRL_C_POPEN_SHORT,
    SECOND_SLEEP_FOR_TEST_MIDDLE,
    SECOND_SLEEP_FOR_TEST_SHORT,
)
from tests.testlibraries.cpu_bound import expect_process_cpu_bound, process_cpu_bound
from tests.testlibraries.local_socket import LocalSocket

if sys.platform == "win32":
    # Reason pylint issue. When put into group, wrong-import-position occur. pylint: disable=ungrouped-imports
    from subprocess import CREATE_NEW_PROCESS_GROUP


async def keyboard_interrupt() -> None:
    await process_cpu_bound()
    _thread.interrupt_main()


class TestRun:
    """Test for run()."""

    @staticmethod
    def test_run() -> None:
        expect = expect_process_cpu_bound(1)
        actual = run(process_cpu_bound, 1)
        assert actual == expect

    @staticmethod
    def test_run_keyboard_interrupt() -> None:
        loop = asyncio.new_event_loop()
        with ProcessPoolExecutor() as executor:
            future = cast(Future, loop.run_in_executor(executor, run, keyboard_interrupt))
        assert not future.get_loop().is_running()
        assert not future.done()

    def test_run_terminate(self) -> None:
        """Function run() should stop when send signal for terminate to child processes."""
        self.execute_test_run(self.terminate)

    def test_run_kill(self) -> None:
        """Function run() should stop when send signal for kill to child processes."""
        self.execute_test_run(self.kill)

    @classmethod
    def execute_test_run(cls, send_signal: Callable[[Process], None]):
        """Executes test run"""
        expect = expect_process_cpu_bound(1)
        future = cls.run_in_process_executor(1)
        cls.execute_send_signal(send_signal)
        loop = future.get_loop()
        assert not loop.is_running()
        loop.run_until_complete(asyncio.wait_for(future, SECOND_SLEEP_FOR_TEST_MIDDLE))
        assert future.done()
        assert future.result() == expect
        assert future.exception() is None

    @staticmethod
    def execute_send_signal(send_signal: Callable[[Process], None]):
        pytest_process = psutil.Process(os.getpid())
        children = pytest_process.children()
        for process in children:
            send_signal(process)

    @staticmethod
    def run_in_process_executor(task_id: Any) -> Future:
        """Sends signal for test."""
        loop = asyncio.new_event_loop()
        with ProcessPoolExecutor() as executor:
            return cast(Future, loop.run_in_executor(executor, run, process_cpu_bound, task_id))

    @staticmethod
    def terminate(process: Process) -> None:
        process.terminate()

    @staticmethod
    def kill(process: Process) -> None:
        process.kill()


class TestTerminateProcess:
    """Test for terminate_process()."""

    def test_terminate_processes(self) -> None:
        """Function terminate_process() should terminate all child processes."""

        def execute_terminate_processes(process_id: int):
            terminate_processes(process_id)

        self.execute_test_terminate_processes(execute_terminate_processes)

    def test_terminate_processes_force(self) -> None:
        """Function terminate_process() should kill all child processes."""

        def execute_terminate_processes(process_id: int):
            terminate_processes(process_id, force=True)

        self.execute_test_terminate_processes(execute_terminate_processes)

    @classmethod
    def execute_test_terminate_processes(cls, callable_execute_terminate_processes: Callable[[int], None]):
        """Executes test for terminate_process()."""
        child_process = cls.create_child_process()
        grandchildren = cls.get_grandchildren_process(child_process)
        assert child_process.pid is not None
        callable_execute_terminate_processes(child_process.pid)
        multiprocessing.connection.wait([child_process.sentinel])
        cls.assert_that_descendant_processes_are_terminated(child_process, grandchildren)

    @classmethod
    def create_child_process(cls) -> Process:
        """Creates child process for test."""
        event = multiprocessing.Event()
        child_process = Process(target=cls.child, args=(event,))
        child_process.start()
        assert child_process.is_alive()
        event.wait()
        return child_process

    @staticmethod
    def get_grandchildren_process(child_process: Process) -> List[psutil.Process]:
        """Creates grandchildren processes for test."""
        psutil_child_process = psutil.Process(child_process.pid)
        grandchildren = psutil_child_process.children(recursive=True)
        assert grandchildren
        for grandchild in grandchildren:
            assert grandchild.is_running()
        return grandchildren

    @classmethod
    def child(cls, event: threading.Event) -> None:
        Process(target=cls.process_cpu_bound).start()
        event.set()

    @staticmethod
    def process_cpu_bound() -> None:
        asyncio.run(process_cpu_bound())

    @staticmethod
    def assert_that_descendant_processes_are_terminated(child_process: Process, grandchildren: List[psutil.Process]):
        _, alive = psutil.wait_procs(grandchildren, timeout=1)
        assert not alive
        assert not child_process.is_alive()
        for grandchild in grandchildren:
            assert not grandchild.is_running()


class TestProcessTaskPoolExecutor:
    """Test for process_task_pool_executor."""

    LOGGER = getLogger(__name__)

    def test_smoke(self) -> None:
        expects = [expect_process_cpu_bound(i) for i in [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]]
        actuals = asyncio.run(self.example_use_case())
        assert actuals is not None
        for expect in expects:
            assert expect in actuals

    # Since Python can't trap signal.SIGTERM in Windows.
    # see:
    #     - Windows: signal doc should state certains signals can't be registered
    #     https://bugs.python.org/issue26350
    @pytest.mark.skipif(sys.platform == "win32", reason="test for Linux only")
    def test_keyboard_interrupt(self) -> None:

        signal.signal(signal.SIGTERM, self.disable_termination)
        asyncio.run(self.keyboard_interrupt_terminate())

    # Since Python can't trap signal.SIGTERM in Windows.
    # see:
    #     - Windows: signal doc should state certains signals can't be registered
    #     https://bugs.python.org/issue26350
    @pytest.mark.skipif(sys.platform == "win32", reason="test for Linux only")
    def test_keyboard_interrupt_sigterm(self) -> None:
        signal.signal(signal.SIGTERM, self.disable_termination)
        asyncio.run(self.keyboard_interrupt_sigterm())

    @staticmethod
    def disable_termination(_signum, _frame):
        pass

    # Since only SIGTERM, CTRL_C_EVENT and CTRL_BREAK_EVENT signals are supported on Windows.
    # see: https://github.com/giampaolo/psutil/blob/e80cabe5206fd7ef14fd6a47e2571f660f95babf/psutil/_pswindows.py#L875
    @pytest.mark.skipif(sys.platform == "win32", reason="test for Linux only")
    def test_keyboard_interrupt_on_linux(self) -> None:
        with pytest.raises(KeyboardInterrupt):
            asyncio.run(self.keyboard_interrupt(signal.SIGINT))

    @staticmethod
    @pytest.mark.skipif(sys.platform != "win32", reason="test for Windows only")
    def test_keyboard_interrupt_ctrl_c_new_window() -> None:
        """
        see:
          - Answer: Sending ^C to Python subprocess objects on Windows
            https://stackoverflow.com/a/7980368/12721873
        """
        popen = Popen(f"start {sys.executable} tests\\testlibraries\\subprocess_wrapper_windows.py", shell=True)
        assert LocalSocket.receive() == "Test succeed"
        assert popen.wait() == 0

    @staticmethod
    @pytest.mark.skipif(sys.platform != "win32", reason="test for Windows only")
    def test_keyboard_interrupt_ctrl_c_popen():
        """
        see:
          - On Windows, what is the python launcher 'py' doing that lets control-C cross between process groups?
            https://stackoverflow.com/q/42180468/12721873
            https://github.com/njsmith/appveyor-ctrl-c-test/blob/34e13fab9be56d59c3eba566e26d80505c309438/a.py
            https://github.com/njsmith/appveyor-ctrl-c-test/blob/34e13fab9be56d59c3eba566e26d80505c309438/run-a.py
        """
        popen = Popen(
            f"{sys.executable} tests\\testlibraries\\keyboaard_interrupt_in_windows.py",
            creationflags=CREATE_NEW_PROCESS_GROUP,
        )
        asyncio.run(asyncio.sleep(SECOND_SLEEP_FOR_TEST_KEYBOARD_INTERRUPT_CTRL_C_POPEN_SHORT))
        LocalSocket.send(str(popen.pid))
        assert LocalSocket.receive() == "Test succeed"
        assert popen.wait() == 0

    @classmethod
    async def keyboard_interrupt_terminate(cls) -> None:
        """
        Can't test in case process.kill() since it sends signal.SIGKILL and Python can't trap it.
        Function process.kill() stops pytest process.
        see:
          - https://psutil.readthedocs.io/en/latest/#psutil.Process.terminate
          - https://psutil.readthedocs.io/en/latest/#psutil.Process.kill
          - https://docs.python.org/ja/3/library/signal.html#signal.SIGKILL
        """
        task = asyncio.create_task(cls.example_use_case_cancel())
        await asyncio.sleep(SECOND_SLEEP_FOR_TEST_SHORT)
        current_process = psutil.Process(os.getpid())
        try:
            current_process.terminate()
        except CancelledError:
            await task
            assert task.done()
            assert isinstance(task.exception(), KeyboardInterrupt)

    @classmethod
    async def keyboard_interrupt_sigterm(cls) -> None:
        """Simulates keyboard interrupt by SIGTERM."""
        task = asyncio.create_task(cls.example_use_case_cancel())
        await asyncio.sleep(SECOND_SLEEP_FOR_TEST_SHORT)
        current_process = psutil.Process(os.getpid())
        try:
            current_process.send_signal(signal.SIGTERM)
        except CancelledError:
            await task
            assert task.done()
            assert task.exception() is None

    @classmethod
    async def keyboard_interrupt(cls, signal_number: int, pid=None) -> None:
        """Simulates keyboard interrupt."""
        if pid is None:
            pid = os.getpid()
        task = asyncio.create_task(cls.example_use_case_cancel())
        await asyncio.sleep(SECOND_SLEEP_FOR_TEST_SHORT)
        current_process = psutil.Process(pid)
        try:
            current_process.send_signal(signal_number)
            await task
        except CancelledError:
            await task
            assert task.done()
            assert task.exception() is None

    @staticmethod
    async def example_use_case() -> List[Future]:
        """The example use case of ProcessTaskPoolExecutor for E2E testing."""
        with ProcessTaskPoolExecutor(max_workers=3, cancel_tasks_when_shutdown=True) as executor:
            futures = {executor.create_process_task(process_cpu_bound, x) for x in range(10)}
            return await asyncio.gather(*futures)

    @staticmethod
    async def example_use_case_cancel() -> None:
        """The example use case of ProcessTaskPoolExecutor for E2E testing in case of cancel."""
        with ProcessTaskPoolExecutor(max_workers=3, cancel_tasks_when_shutdown=True) as executor:
            futures = {executor.create_process_task(process_cpu_bound, x) for x in range(10)}
            await asyncio.gather(*futures)
            pytest.fail()



"""Top-level package for Asynchronous CPU."""
from typing import List

from asynccpu.process_task_pool_executor import *  # noqa

__author__ = """Yukihiko Shinoda"""
__email__ = "yuk.hik.future@gmail.com"
__version__ = "1.0.0"

__all__: List[str] = []
# pylint: disable=undefined-variable
__all__ += process_task_pool_executor.__all__  # type: ignore # noqa



"""The example use case of ProcessTaskPoolExecutor for E2E testing in case of cancel."""
import asyncio

from asynccpu import ProcessTaskPoolExecutor
from tests.testlibraries.cpu_bound import process_cpu_bound


async def example_use_case_cancel() -> None:
    """The example use case of ProcessTaskPoolExecutor for E2E testing in case of cancel."""
    with ProcessTaskPoolExecutor(max_workers=3, cancel_tasks_when_shutdown=True) as executor:
        futures = {executor.create_process_task(process_cpu_bound, x) for x in range(10)}
        await asyncio.gather(*futures)
        raise Exception("Failed")



"""Configuration of pytest"""
import pytest  # type: ignore

collect_ignore = ["setup.py"]


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')



"""Test libraries."""
SECOND_SLEEP_FOR_TEST_SHORT = 0.12
SECOND_SLEEP_FOR_TEST_MIDDLE = 0.25
# When set 150, test_keyboard_interrupt_ctrl_c_popen failed in GitHub Actions.
SECOND_SLEEP_FOR_TEST_KEYBOARD_INTERRUPT_CTRL_C_POPEN_SHORT = 180
SECOND_SLEEP_FOR_TEST_KEYBOARD_INTERRUPT_CTRL_C_POPEN_MIDDLE = 180



"""The socket to comunicate processes of different group each other."""
import socket
from contextlib import closing

from tests.testlibraries import SECOND_SLEEP_FOR_TEST_KEYBOARD_INTERRUPT_CTRL_C_POPEN_MIDDLE


class LocalSocket:
    """The socket to comunicate processes of different group each other."""

    ENCODING = "utf-8"
    HOST = "127.0.0.1"
    # Almost ports are blocked on GitHub Windows runner.
    # Although the port 8080 seems to be able to use.
    # see: https://github.community/t/what-is-the-default-firewall-for-github-actions-runners/17732
    PORT = 8080
    TIME_TO_ALLOW_ACCESS = 1
    TIMEOUT = SECOND_SLEEP_FOR_TEST_KEYBOARD_INTERRUPT_CTRL_C_POPEN_MIDDLE
    MAX_BUFFER_SIZE = 1024
    SOCKET_REUSE_ADDRESS_ENABLE = 1

    @staticmethod
    def send(message: str):
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as socket_to:
            socket_to.connect((LocalSocket.HOST, LocalSocket.PORT))
            socket_to.send(bytes(message, LocalSocket.ENCODING))

    @staticmethod
    def receive():
        """Receives message."""
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as socket_from:
            socket_from.settimeout(LocalSocket.TIMEOUT)
            # see:
            #   - https://docs.python.org/ja/3/library/socket.html#example
            #   - Answer: Python server “Only one usage of each socket address is normally permitted”
            #     https://stackoverflow.com/a/12362623/12721873
            socket_from.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, LocalSocket.SOCKET_REUSE_ADDRESS_ENABLE)
            socket_from.bind((LocalSocket.HOST, LocalSocket.PORT))
            socket_from.listen(LocalSocket.TIME_TO_ALLOW_ACCESS)
            connection, _ = socket_from.accept()
            try:
                return str(connection.recv(LocalSocket.MAX_BUFFER_SIZE), LocalSocket.ENCODING)
            finally:
                connection.close()



"""Test of keyboard Interrupt for Windows."""
# Reason: only for Windows. pylint: disable=import-error
import win32api  # type: ignore

from tests.testlibraries.keyboard_interrupter import KeyboardInterrupter
from tests.testlibraries.local_socket import LocalSocket

win32api.SetConsoleCtrlHandler(None, False)


async def get_process_id():
    print("Await socket")
    process_id = int(LocalSocket.receive())
    print("Kill group lead process")
    return process_id


if __name__ == "__main__":
    try:
        KeyboardInterrupter(get_process_id()).test_keyboard_interrupt()
    except KeyboardInterrupt:
        pass



"""Subprocess wrapper for Windows."""
import time

from tests.testlibraries import SECOND_SLEEP_FOR_TEST_SHORT
from tests.testlibraries.keyboard_interrupter import KeyboardInterrupter


async def get_process_id():
    print("Await sleep")
    time.sleep(SECOND_SLEEP_FOR_TEST_SHORT)
    print("Kill all processes in this window.")
    return 0


if __name__ == "__main__":
    try:
        KeyboardInterrupter(get_process_id()).test_keyboard_interrupt()
    except KeyboardInterrupt:
        time.sleep(10)



"""
Cpu bound.
see: https://docs.python.org/3/library/asyncio-eventloop.html#executing-code-in-thread-or-process-pools
"""


async def process_cpu_bound(task_id: int = None):
    """
    CPU-bound operations will block the event loop:
    in general it is preferable to run them in a process pool.
    """
    return ("" if task_id is None else f"task_id: {task_id}, ") + f"result: {sum(i * i for i in range(10 ** 7))}"


def expect_process_cpu_bound(task_id: int = None):
    return ("" if task_id is None else f"task_id: {task_id}, ") + "result: 333333283333335000000"



"""Process task pool executor."""
from __future__ import annotations

import asyncio
import os
import sys
import threading
import weakref
from asyncio.futures import Future
from concurrent.futures import ProcessPoolExecutor
from logging import getLogger
from typing import Awaitable, Callable, Literal, cast

import psutil

__all__ = ["ProcessTaskPoolExecutor"]


def run(corofn: Callable[..., Awaitable], *args):
    """
    This function requires to be orphaned since ProcecssPoolExecutor requires picklable.
    The argument of Coroutine requires not "raw Corutine object" but "Coroutine function"
    since raw Coroutine object is not picklable:
      TypeError: cannot pickle 'coroutine' object
    see:
      - multiprocessing — Process-based parallelism
        https://docs.python.org/3/library/multiprocessing.html
      - Answer: Python multiprocessing PicklingError: Can't pickle <type 'function'> - Stack Overflow
        https://stackoverflow.com/a/8805244/12721873
      - Answer: Why coroutines cannot be used with run_in_executor? - Stack Overflow
        https://stackoverflow.com/a/46075571/12721873
    """
    logger = getLogger(__name__)
    logger.debug("Start run coroutine")
    loop = asyncio.new_event_loop()
    try:
        coro = corofn(*args)
        asyncio.set_event_loop(loop)
        return loop.run_until_complete(coro)
    except KeyboardInterrupt:  # pragma: no cover
        logger.info("Cancel run coroutine")
        # Since this process takes next coroutine and execute it
        # even if succeed to cancel current coroutine.
        # see:
        #   - Answer: multiprocess - Python: concurrent.futures How to make it cancelable? - Stack Overflow
        #     https://stackoverflow.com/a/45515052/12721873
        terminate_processes(os.getpid())
        raise
    finally:
        loop.close()
        logger.debug("Finish run coroutine")


def terminate_processes(parent_pid: int, *, force: bool = False):
    """
    This method doesn't have parameter for sending signal
    since psutil seems to being blackbox the difference between Linux and Windows.
    see:
      - Answer: multiprocess - Python: concurrent.futures How to make it cancelable? - Stack Overflow
        https://stackoverflow.com/a/45515052/12721873
    """
    logger = getLogger(__name__)
    parent = psutil.Process(parent_pid)
    children = parent.children(recursive=True)
    logger.debug("Terminate child processes")
    if force:
        for process in children:
            process.kill()
        logger.debug("Terminate target processes")
        parent.kill()
    else:
        for process in children:
            process.terminate()
        logger.debug("Terminate target processes")
        parent.terminate()


class ProcessTaskPoolExecutor(ProcessPoolExecutor):
    """
    see:
      - Answer: What's the correct way to clean up after an interrupted event loop?
        https://stackoverflow.com/a/30766124/12721873
        https://stackoverflow.com/a/58532304/12721873
    """

    def __init__(
        self,
        max_workers=None,
        mp_context=None,
        initializer=None,
        initargs=(),
        *,
        cancel_tasks_when_shutdown: bool = False
    ) -> None:
        super().__init__(max_workers, mp_context, initializer, initargs)
        self.cancel_tasks_when_shutdown = cancel_tasks_when_shutdown
        self.loop = asyncio.get_event_loop()
        self.list_process_task: weakref.WeakSet[Future] = weakref.WeakSet()
        self.lock_for_list_process_task = threading.Lock()
        self.logger = getLogger(__name__)

    def __enter__(self) -> ProcessTaskPoolExecutor:
        # pylint can't detect following cast:
        # return cast(ProcessTaskPoolExecutor, super().__enter__())
        return self

    # Reason: pylint bug. pylint: disable=unsubscriptable-object
    def __exit__(self, exc_type, exc_val, exc_tb) -> Literal[False]:
        self.logger.debug("Exit ProcessTaskPoolExecutor: Start")
        self.logger.debug("exc_type = %s", exc_type)
        self.logger.debug("Exit ProcessTaskPoolExecutor: Cancelled all process tasks")
        with self.lock_for_list_process_task:
            for task in self.list_process_task:
                self.cancel_if_not_cancelled(task)
        self.shutdown()
        self.logger.debug("Exit ProcessTaskPoolExecutor: Finish")
        return False

    def shutdown(self, wait=True, *, cancel_futures=None):
        if cancel_futures is None:
            cancel_futures = self.cancel_tasks_when_shutdown
        if sys.version_info.major < 3 or sys.version_info.major == 3 and sys.version_info.minor <= 8:
            super().shutdown(wait)  # pragma: no cover
        else:
            # Reason: mypy bug
            super().shutdown(wait, cancel_futures=cancel_futures)  # type: ignore

    def cancel_if_not_cancelled(self, task: Future) -> None:
        """Cancels task like future if its not cancelled."""
        self.logger.debug(task)
        cancelled = task.cancelled()
        self.logger.debug("Cancelled?: %s", cancelled)
        if cancelled:
            return
        cancel = task.cancel()
        self.logger.debug("Cancel succeed?: %s", cancel)

    def create_process_task(self, function_coroutine, *args) -> Future:
        """Creates task like future by wraping coroutine."""
        with self.lock_for_list_process_task:
            task = cast(Future, self.loop.run_in_executor(self, run, function_coroutine, *args))
            self.list_process_task.add(task)
            self.list_process_task = self.garbage_collect(self.list_process_task)
            return task

    @staticmethod
    def garbage_collect(list_process_task: weakref.WeakSet) -> weakref.WeakSet:
        new_list_process_task: weakref.WeakSet[Future] = weakref.WeakSet()
        for process_task in list(list_process_task):
            if not process_task.done():
                new_list_process_task.add(process_task)
        return new_list_process_task



#!/usr/bin/env python
"""The setup script."""

from setuptools import find_packages, setup  # type: ignore

with open("README.md", encoding="utf-8") as readme_file:
    readme = readme_file.read()

# Since new process created by multiprocessing calls setuptools.setup method
# when the test is executed via setup.py and this affects something.
# see:
#   - Answer: How to get `setup.py test` working with `multiprocessing` on Windows?
#     https://stackoverflow.com/a/50402381/12721873
if __name__ == "__main__":
    setup(
        author="Yukihiko Shinoda",
        author_email="yuk.hik.future@gmail.com",
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Framework :: AsyncIO",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Natural Language :: English",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Topic :: Software Development :: Libraries",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Typing :: Typed",
        ],
        dependency_links=[],
        description="Supports async / await pattern for CPU-bound operations.",  # noqa: E501 pylint: disable=line-too-long
        exclude_package_data={"": ["__pycache__", "*.py[co]", ".pytest_cache"]},
        include_package_data=True,
        install_requires=["psutil"],
        keywords="asynccpu",
        long_description=readme,
        long_description_content_type="text/markdown",
        name="asynccpu",
        packages=find_packages(
            include=["asynccpu", "asynccpu.*", "tests", "tests.*"]
        ),  # noqa: E501 pylint: disable=line-too-long
        package_data={"asynccpu": ["py.typed"], "tests": ["*"]},
        python_requires=">=3.8",
        test_suite="tests",
        tests_require=["pytest>=3"],
        url="https://github.com/yukihiko-shinoda/asynccpu",
        version="1.0.0",
        zip_safe=False,
    )



"""To keep task property even raise KeyboardInterrupt."""
import asyncio
import os
import signal
import time
from asyncio.tasks import Task
from typing import Awaitable, Optional

from tests.testlibraries import SECOND_SLEEP_FOR_TEST_MIDDLE
from tests.testlibraries.example_use_case import example_use_case_cancel
from tests.testlibraries.local_socket import LocalSocket


class KeyboardInterrupter:
    """To keep task property even raise KeyboardInterrupt."""

    def __init__(self, get_process_id: Awaitable[int]) -> None:
        # Reason: pytest bug. pylint: disable=unsubscriptable-object
        self.task: Optional[Task] = None
        self.get_process_id = get_process_id

    def test_keyboard_interrupt(self) -> None:
        """Tests keyboard interrupt and send response to pytest by socket when succeed."""
        try:
            asyncio.run(self.keyboard_interrupt())
        except KeyboardInterrupt:
            print("Sleep in except")
            time.sleep(SECOND_SLEEP_FOR_TEST_MIDDLE)
            print("Assert in except")
            assert self.task is not None
            assert self.task.done()
            assert self.task.cancelled()
            LocalSocket.send("Test succeed")
            raise

    async def keyboard_interrupt(self) -> None:
        """Simulates keyboard interrupt by CTRL_C_EVENT."""
        print("Create task")
        self.task = asyncio.create_task(example_use_case_cancel())
        process_id = await self.get_process_id
        try:
            # Reason: only for Windows. pylint: disable=no-member
            os.kill(process_id, signal.CTRL_C_EVENT)  # type: ignore
            print("Await task")
            await self.task
        except KeyboardInterrupt:
            print("Await task in except")
            # await self.task
            print("Assert")
            assert not self.task.done()
            print("Task not done")
            assert not self.task.cancelled()
            print("Task cancelled")
            raise
